//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Written by Philippe Leefsma - Autodesk Developer Technical Services. 2010.
// 
// Feedback and bug report: Philippe.Leefsma@Autodesk.com
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#include "StdAfx.h"
#include <sstream>
#include "AssocSurfacesUtil.h"
#include "AcConstrainedGeometry.h"
#include "ConstrainDbgHelper.h"
#include "AcDbAssocActionBody.h"
#include "AcDbAssocDependency.h"
#include "AcDbAssocPathActionParam.h"
#include "AcDbAssocParamBasedActionBody.h"
#include "AcDbAssocPathBasedSurfaceActionBody.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CAssocSurfacesUtil::CAssocSurfacesUtil(void)
{

}

CAssocSurfacesUtil::~CAssocSurfacesUtil(void)
{

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Use: Get all surface ObjectIds in a given BlockTableRecord
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CAssocSurfacesUtil::getAllSurfaces(const AcDbObjectId& spaceId, vector<AcDbObjectId>& surfaceIds)
{
	AcDbObjectPointer <AcDbBlockTableRecord> btr(spaceId, kForRead);

	AcDbBlockTableRecordIterator* btrIterator;

	btr->newIterator(btrIterator);

	for(btrIterator->start (); !btrIterator->done () ; btrIterator->step ()) 
	{
		AcDbObjectId id;
		btrIterator->getEntityId(id);
		AcDbObjectPointer <AcDbEntity> entity(id, kForRead);

		if(entity->isKindOf(AcDbSurface::desc()))
		{
			surfaceIds.push_back(id);
		}
	}

	delete btrIterator;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Use: Get properties for a given AcDbObject. returns them in a Name-Value pair vector.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CAssocSurfacesUtil::getProperties(const AcDbObjectId& id, vector <ObjectProperty>& properties)
{
	AcDbObjectPointer <AcDbObject> object(id, AcDb::kForRead);

	AcDbObject* obj = object.object();

    CString str;

	properties.push_back(ObjectProperty(L"------ AcDbObject ------"));
	properties.push_back(ObjectProperty(L"Class Name", (ACHAR*)CConstrainDbgHelper::objToClassStr(obj)));
	properties.push_back(ObjectProperty(L"Entity Name", (ACHAR*)CConstrainDbgHelper::objToEnameStr(obj, str)));
	properties.push_back(ObjectProperty(L"Handle", (ACHAR*)CConstrainDbgHelper::objToHandleStr(obj, str)));
	properties.push_back(ObjectProperty(L"Database", (ACHAR*)CConstrainDbgHelper::dbToStr(obj->database(), str)));

    AcDbObject* ownerObj;
    Acad::ErrorStatus es = acdbOpenObject(ownerObj, obj->ownerId(), AcDb::kForRead);
    if (es == Acad::eOk)
	{
        properties.push_back(ObjectProperty(L"Owner", (ACHAR*)CConstrainDbgHelper::objToClassAndHandleStr(ownerObj, str)));
        ownerObj->close();
    }
    else
	{
		properties.push_back(ObjectProperty(L"Owner", (ACHAR*)CConstrainDbgHelper::rxErrorStr(es)));
	}

	properties.push_back(ObjectProperty(L"Merge Style", (ACHAR*)CConstrainDbgHelper::mergeStyleToStr(obj->mergeStyle(), str)));
	properties.push_back(ObjectProperty(L"Has Save Version Override", (ACHAR*)CConstrainDbgHelper::booleanToStr(obj->hasSaveVersionOverride(), str)));

    AcDb::AcDbDwgVersion dwgVer;
    AcDb::MaintenanceReleaseVersion maintVer;
    es = obj->getObjectBirthVersion(dwgVer, maintVer);
    
	if (es == Acad::eOk) 
	{
		properties.push_back(ObjectProperty(L"Birth DWG Version", (ACHAR*)CConstrainDbgHelper::dwgVersionToStr(dwgVer, str)));
		properties.push_back(ObjectProperty(L"Birth Maintenance Version", (ACHAR*)CConstrainDbgHelper::intToStr(maintVer, str)));
    }
    else
	{
		properties.push_back(ObjectProperty(L"Birth Version", (ACHAR*)CConstrainDbgHelper::rxErrorStr(es)));
	} 
	
	if(object->isKindOf(AcDbSurface::desc()))
	{
		AcDbSurface* surface = AcDbSurface::cast(object);

		getSurfaceProperties(surface, properties);

		return;
	}

	if(object->isKindOf(AcDbAssocActionBody::desc()))
	{
		AcDbAssocActionBody* assocBody = AcDbAssocActionBody::cast(object);

		getAssocActionBodyProperties(assocBody, properties);

		return;
	}

	if(object->isKindOf(AcDbAssocPathActionParam::desc()))
	{
		AcDbAssocPathActionParam* pathActionParam = AcDbAssocPathActionParam::cast(object);

		getAssocPathActionParamProperties(pathActionParam, properties);

		return;
	}

	if(object->isKindOf(AcDbAssocActionParam::desc()))
	{
		AcDbAssocActionParam* actionParam = AcDbAssocActionParam::cast(object);

		getAssocActionParamProperties(actionParam, properties);

		return;
	}

	if(object->isKindOf(AcDbAssocDependency::desc()))
	{
		AcDbAssocDependency* dependency = AcDbAssocDependency::cast(object);

		getDependencyProperties(dependency, properties);

		return;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Use: Get properties for a non db-resident entity.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CAssocSurfacesUtil::getProperties(ItemReference* pItemRef, vector <ObjectProperty>& properties)
{
	switch(pItemRef->referenceType)
	{
	case ItemReference::ItemReferenceType::kAcDbEdgeRefReference:

		getEdgeRefProperties((EdgeRefReference*)pItemRef, properties);
		break;

	default:
		break;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
std::wstring CAssocSurfacesUtil::lineWeightToStr(AcDb::LineWeight lineWeight)
{
	switch(lineWeight)
	{
	case AcDb::kLnWt000: return L"LnWt000";        
	case AcDb::kLnWt005: return L"LnWt005";        
	case AcDb::kLnWt009: return L"LnWt009";        
	case AcDb::kLnWt013: return L"LnWt013";        
	case AcDb::kLnWt015: return L"LnWt015";        
	case AcDb::kLnWt018: return L"LnWt018";       
	case AcDb::kLnWt020: return L"LnWt020";        
	case AcDb::kLnWt025: return L"LnWt025";        
	case AcDb::kLnWt030: return L"LnWt030";       
	case AcDb::kLnWt035: return L"LnWt035";       
	case AcDb::kLnWt040: return L"LnWt040";       
	case AcDb::kLnWt050: return L"LnWt050";       
	case AcDb::kLnWt053: return L"LnWt053";       
	case AcDb::kLnWt060: return L"LnWt060";       
	case AcDb::kLnWt070: return L"LnWt070";       
	case AcDb::kLnWt080: return L"LnWt080";       
	case AcDb::kLnWt090: return L"LnWt090";       
	case AcDb::kLnWt100: return L"LnWt100";       
	case AcDb::kLnWt106: return L"LnWt106";       
	case AcDb::kLnWt120: return L"LnWt120";       
	case AcDb::kLnWt140: return L"LnWt140";       
	case AcDb::kLnWt158: return L"LnWt158";      
	case AcDb::kLnWt200: return L"LnWt200";      
	case AcDb::kLnWt211: return L"LnWt211";       
	case AcDb::kLnWtByLayer: return L"ByLayer";   
	case AcDb::kLnWtByBlock: return L"ByBlock";  
	case AcDb::kLnWtByLwDefault: return L"ByDefault";
        
	default:
		return L"Unknown";
	}
}          
               
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CAssocSurfacesUtil::getSurfaceProperties(AcDbSurface* surface, vector <ObjectProperty>& properties)
{
	CString str;

	properties.push_back(ObjectProperty());
	properties.push_back(ObjectProperty(L"------ AcDbSurface ------"));

	AcDbObjectId actionBodyId = surface->creationActionBodyId();

	bool isAssociative = (actionBodyId.isNull() ? false : true);
	properties.push_back(ObjectProperty(L"Is Associative", (ACHAR*)CConstrainDbgHelper::booleanToStr(isAssociative, str)));

	AcDbExtents bounds;
	bool ret = surface->bounds(bounds);

	bool isPlanar = surface->isPlanar();
	properties.push_back(ObjectProperty(L"Is Planar", (ACHAR*)CConstrainDbgHelper::booleanToStr(isPlanar, str)));

	ACHAR* layer = surface->layer();
	properties.push_back(ObjectProperty(L"Layer", layer));
	acutDelString(layer);

	ACHAR* material = surface->material();
	properties.push_back(ObjectProperty(L"Material", material));
	acutDelString(material);

	ACHAR* linetype = surface->linetype();
	properties.push_back(ObjectProperty(L"Line Type", linetype));
	acutDelString(linetype);

	double linetypeScale = surface->linetypeScale();
	properties.push_back(ObjectProperty(L"Line Type Scale", (ACHAR*)CConstrainDbgHelper::doubleToStr(linetypeScale, str)));
	
	AcDb::LineWeight lineWeight = surface->lineWeight();
	properties.push_back(ObjectProperty(L"Line Weight", lineWeightToStr(lineWeight)));
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
std::wstring StringToWString(const std::string& s)
{
	std::wstring temp(s.length(),L' ');
	std::copy(s.begin(), s.end(), temp.begin());
	return temp;
}

std::wstring numericToWString(const int& in) 
{
	std::ostringstream ost(ostringstream::out);
	ost << in;

	return StringToWString(ost.str());
}

std::wstring numericToWString(const long& in) 
{
	std::ostringstream ost(ostringstream::out);
	ost << in;

	return StringToWString(ost.str());
}

std::wstring numericToWString(const double& in) 
{
	std::ostringstream ost(ostringstream::out);
	ost << in;

	return StringToWString(ost.str());
}

std::wstring CAssocSurfacesUtil::variantToStr(const AcDbEvalVariant& eval)
{
	AcDb::DwgDataType propType = eval.getType();

	switch(propType)
	{
	case AcDb::DwgDataType::kDwgNull:
		return L"None";

	case AcDb::DwgDataType::kDwgReal:
		return numericToWString(eval.resval.rreal);
	
	case AcDb::DwgDataType::kDwgInt16:
		return numericToWString(eval.resval.rint);

	case AcDb::DwgDataType::kDwgInt32:
		return numericToWString(eval.resval.rint);
		
	case AcDb::DwgDataType::kDwgInt64:
		return numericToWString(eval.resval.rlong);
		
	case AcDb::DwgDataType::kDwg3Real:
	
		if(eval.restype == AcDb::kDxfXCoord)
		{
			std::wstring res;
			res += L"[" + numericToWString(eval.resval.rpoint[0]) + L", ";
			res += numericToWString(eval.resval.rpoint[1]) + L"]";
			return res;
		}

		if(eval.restype == AcDb::kDxfXCoord+1)
		{
			std::wstring res;
			res += L"[" + numericToWString(eval.resval.rpoint[0]) + L", ";
			res += numericToWString(eval.resval.rpoint[1]) + L", ";
			res += numericToWString(eval.resval.rpoint[2]) + L"]";
			return res;
		}
		
		break;

	case AcDb::DwgDataType::kDwgText:
		return eval.resval.rstring;

	default:
		break;
	}

	return L"";
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CAssocSurfacesUtil::getAssocActionBodyProperties(AcDbAssocActionBody* actionBody, vector <ObjectProperty>& properties)
{
	CString str;

	properties.push_back(ObjectProperty());
	properties.push_back(ObjectProperty(L"------ AcDbAssocActionBody ------"));

	AcDbObject* parentAction;
    Acad::ErrorStatus es = acdbOpenObject(parentAction, actionBody->parentAction(), AcDb::kForRead);
    if (es == Acad::eOk)
	{
        properties.push_back(ObjectProperty(L"Parent Action", (ACHAR*)CConstrainDbgHelper::objToClassAndHandleStr(parentAction, str)));
        parentAction->close();
    }
    else
	{
		properties.push_back(ObjectProperty(L"Parent Action", (ACHAR*)CConstrainDbgHelper::rxErrorStr(es)));
	}

	if (actionBody->isKindOf(AcDbAssocParamBasedActionBody::desc()))
	{
		AcDbAssocParamBasedActionBody* paramBasedActionBody = AcDbAssocParamBasedActionBody::cast(actionBody);

		AcArray<AcString> paramNames;
		paramBasedActionBody->ownedValueParamNames(paramNames);

		properties.push_back(ObjectProperty());
		properties.push_back(ObjectProperty(L"Parameters: "));

		for(int i = 0; i<paramNames.length(); ++i)
		{
			AcDbEvalVariant value;
			AcString name = paramNames[i];
			AcString expression;
			AcString evaluatorId;
			
			if((es = paramBasedActionBody->getValueParam(name, value, expression, evaluatorId)) == Acad::eOk)
			{
				std::wstring wsname = name.constPtr();
				std::wstring wsexpression = expression.constPtr();
				
				//No expression, need to get the value
				if(!wsexpression.length())
				{
					wsexpression = variantToStr(value);
				}

				properties.push_back(ObjectProperty(wsname, wsexpression));
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CAssocSurfacesUtil::getAssocPathActionParamProperties(AcDbAssocPathActionParam* assocPathActionParam, vector <ObjectProperty>& properties)
{
	CString str;

	properties.push_back(ObjectProperty());
	properties.push_back(ObjectProperty(L"------ AcDbAssocPathActionParam ------"));

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CAssocSurfacesUtil::getAssocActionParamProperties(AcDbAssocActionParam* assocActionParam, vector <ObjectProperty>& properties)
{
	CString str;

	properties.push_back(ObjectProperty());
	properties.push_back(ObjectProperty(L"------ AcDbAssocActionParam ------"));

}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
std::wstring CAssocSurfacesUtil::curveTypeToStr(AcGe::EntityId type)
{
	switch(type)
	{
	case AcGe::kEntity2d:			return L"Entity2d";
	case AcGe::kEntity3d:			return L"Entity3d";
	case AcGe::kPointEnt2d:			return L"PointEnt2d";
	case AcGe::kPointEnt3d:			return L"PointEnt3d";
	case AcGe::kPosition2d:			return L"Position2d";	
	case AcGe::kPosition3d:			return L"Position3d";
	case AcGe::kPointOnCurve2d:		return L"PointOnCurve2d";
	case AcGe::kPointOnCurve3d:		return L"PointOnCurve3d";
	case AcGe::kBoundedPlane:		return L"BoundedPlane";
	case AcGe::kCircArc2d:			return L"CircArc2d";
	case AcGe::kCircArc3d:			return L"CircArc3d";
	case AcGe::kConic2d:			return L"Conic2d";
	case AcGe::kConic3d:			return L"Conic3d";
	case AcGe::kCurve2d:			return L"Curve2d";
	case AcGe::kCurve3d:			return L"Curve3d";
	case AcGe::kEllipArc2d:			return L"EllipArc2d";
	case AcGe::kEllipArc3d:			return L"EllipArc3d";
	case AcGe::kHelix:				return L"Helix";
	case AcGe::kLine2d:				return L"Line2d";
	case AcGe::kLine3d:				return L"Line3d";
	case AcGe::kLinearEnt2d:		return L"LinearEnt2d";
	case AcGe::kLinearEnt3d:		return L"LinearEnt3d";
	case AcGe::kLineSeg2d:			return L"LineSeg2d";
	case AcGe::kLineSeg3d:			return L"LineSeg3d";
	case AcGe::kPlanarEnt:			return L"PlanarEnt";
	case AcGe::kExternalCurve3d:	return L"ExternalCurve3d";
	case AcGe::kExternalCurve2d:	return L"ExternalCurve2d";
	case AcGe::kPlane:				return L"Plane";
	case AcGe::kRay2d:				return L"Ray2d";
	case AcGe::kRay3d:				return L"Ray3d";
	case AcGe::kSurface:			return L"Surface";
	case AcGe::kSphere:				return L"Sphere";
	case AcGe::kCone:				return L"Cone";
	case AcGe::kTorus:				return L"Torus";
	case AcGe::kCylinder:			return L"Cylinder";
	case AcGe::kSplineEnt2d:		return L"SplineEnt2d";
	case AcGe::kSurfaceCurve2dTo3d:	return L"SurfaceCurve2dTo3d";
	case AcGe::kSurfaceCurve3dTo2d:	return L"SurfaceCurve3dTo2d";
	case AcGe::kPolyline2d:			return L"Polyline2d";
	case AcGe::kAugPolyline2d:		return L"AugPolyline2d";
	case AcGe::kNurbCurve2d:		return L"NurbCurve2d";
	case AcGe::kDSpline2d:			return L"DSpline2d";
	case AcGe::kCubicSplineCurve2d:	return L"CubicSplineCurve2d";
	case AcGe::kSplineEnt3d:		return L"SplineEnt3d";
	case AcGe::kPolyline3d:			return L"Polyline3d";
	case AcGe::kAugPolyline3d:		return L"AugPolyline3d";
	case AcGe::kNurbCurve3d:		return L"NurbCurve3d";
	case AcGe::kDSpline3d:			return L"DSpline3d";
	case AcGe::kCubicSplineCurve3d:	return L"CubicSplineCurve3d";
	case AcGe::kTrimmedCrv2d:		return L"TrimmedCrv2d";
	case AcGe::kCompositeCrv2d:		return L"CompositeCrv2d";
	case AcGe::kCompositeCrv3d:		return L"CompositeCrv3d";
	case AcGe::kEnvelope2d:			return L"Envelope2d";
	case AcGe::kExternalSurface:	return L"ExternalSurface";
	case AcGe::kNurbSurface:		return L"NurbSurface";
	case AcGe::kOffsetSurface:		return L"OffsetSurface";
	case AcGe::kTrimmedSurface:		return L"TrimmedSurface";
	case AcGe::kCurveBoundedSurface:return L"CurveBoundedSurface";
	case AcGe::kPointOnSurface:		return L"PointOnSurface";
	case AcGe::kExternalBoundedSurface:return L"ExternalBoundedSurface";
	case AcGe::kCurveCurveInt2d:	return L"CurveCurveInt2d";
	case AcGe::kCurveCurveInt3d:	return L"CurveCurveInt3d";
	case AcGe::kBoundBlock2d:		return L"BoundBlock2d";
	case AcGe::kBoundBlock3d:		return L"BoundBlock3d";
	case AcGe::kOffsetCurve2d:		return L"OffsetCurve2d";
	case AcGe::kOffsetCurve3d:		return L"OffsetCurve3d";
	case AcGe::kPolynomCurve3d:		return L"PolynomCurve3d";
	case AcGe::kBezierCurve3d:		return L"BezierCurve3d";
	case AcGe::kObject:				return L"Object";
	case AcGe::kFitData3d:			return L"FitData3d";
	case AcGe::kHatch:				return L"Hatch";
	case AcGe::kTrimmedCurve2d:		return L"TrimmedCurve2d";	
	case AcGe::kTrimmedCurve3d:		return L"TrimmedCurve3d";
	case AcGe::kCurveSampleData:	return L"CurveSampleData";
	case AcGe::kEllipCone:			return L"EllipCone";
	case AcGe::kEllipCylinder:		return L"EllipCylinder";
	case AcGe::kIntervalBoundBlock:	return L"IntervalBoundBlock";
	case AcGe::kClipBoundary2d:		return L"ClipBoundary2d";
	case AcGe::kExternalObject:		return L"ExternalObject";
	case AcGe::kCurveSurfaceInt:	return L"CurveSurfaceInt";
	case AcGe::kSurfaceSurfaceInt:	return L"SurfaceSurfaceInt";

	default:
		return L"Unknown Type";
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CAssocSurfacesUtil::getEdgeRefProperties(EdgeRefReference* edgeRefReference, vector <ObjectProperty>& properties)
{
	CString str;

	properties.push_back(ObjectProperty(L"------ AcDbEdgeRef ------"));

	AcDbObjectPointer <AcDbAssocPathActionParam> param(edgeRefReference->paramId, AcDb::kForRead);

	if(param.openStatus() != Acad::eOk)
		return;

	AcArray<AcArray<AcDbEdgeRef>> edgeRefs;
	param->getEdgeRefArray(edgeRefs);

	AcDbEdgeRef edgeRef = edgeRefs[edgeRefReference->iIndex][edgeRefReference->jIndex];

	const AcGeCurve3d* curve = edgeRef.curve();
	AcDbEntity* entityRef = edgeRef.createEntity();

	properties.push_back(ObjectProperty(L"Curve Type", curveTypeToStr(curve->type())));
	properties.push_back(ObjectProperty(L"Entity Reference", (ACHAR*)entityRef->isA()->name()));

	delete entityRef;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CAssocSurfacesUtil::getDependencyProperties(AcDbAssocDependency* dependency, vector <ObjectProperty>& properties)
{
	CString str;

	properties.push_back(ObjectProperty());
	properties.push_back(ObjectProperty(L"------ AcDbAssocDependency ------"));

}