﻿// (C) Copyright 2002-20010 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// AcHelper is a utility class for the Surfaces Sample
//
// Products: AutoCAD (and verticals) 2011 and higher
//
// Written by Philippe Leefsma - 2009/2010 - Developer Technical Services
// Philippe.Leefsma@Autodesk.com
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Text;

using Autodesk.AutoCAD;
using Autodesk.AutoCAD.Runtime;
using Autodesk.AutoCAD.EditorInput;
using Autodesk.AutoCAD.DatabaseServices;
using Autodesk.AutoCAD.ApplicationServices;

[assembly: CommandClass(typeof(HammerSurfaces.AcHelper))]

namespace HammerSurfaces
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Helper Class
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    public class AcHelper
    {
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Post an Entity to the specified database model space
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static ObjectId PostToDatabase(Database db, Entity entity)
        {
            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                BlockTable bt = Tx.GetObject(db.BlockTableId, OpenMode.ForRead) as BlockTable;

                BlockTableRecord model = Tx.GetObject(bt[BlockTableRecord.ModelSpace], OpenMode.ForWrite) as BlockTableRecord;

                ObjectId newId = model.AppendEntity(entity);

                Tx.AddNewlyCreatedDBObject(entity, true);

                Tx.Commit();

                return newId;
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Post an Entity to the current document model space
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static ObjectId PostToDatabase(Entity entity)
        {
            Database db = Application.DocumentManager.MdiActiveDocument.Database;

            return PostToDatabase(db, entity);
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Get edges subentities from input entity using Brep API
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static FullSubentityPath[] GetEdgesSubentityBrep(Entity entity)
        {
            Database db = Application.DocumentManager.MdiActiveDocument.Database;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                try
                {
                    FullSubentityPath EntityPath = new FullSubentityPath(new ObjectId[] { entity.ObjectId }, new SubentityId(SubentityType.Null, 0));

                    List<FullSubentityPath> edgesPath = new List<FullSubentityPath>();

                    using (Autodesk.AutoCAD.BoundaryRepresentation.Brep brep = new Autodesk.AutoCAD.BoundaryRepresentation.Brep(EntityPath))
                    {
                        foreach (Autodesk.AutoCAD.BoundaryRepresentation.Edge edge in brep.Edges)
                        {
                            edgesPath.Add(edge.SubentityPath);
                        }
                    }

                    if (edgesPath.Count == 0) 
                        return null;

                    return edgesPath.ToArray();
                }
                catch
                {
                    return null;
                }
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Get edges subentities from input entity using AssocPersSubentityIdPE protocol extension
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static FullSubentityPath[] GetEdgeSubentities(Entity entity)
        {
            ObjectId[] entId = new ObjectId[] { entity.ObjectId };

            IntPtr pSubentityIdPE = entity.QueryX(AssocPersSubentityIdPE.GetClass(typeof(AssocPersSubentityIdPE)));

            if (pSubentityIdPE == IntPtr.Zero) 
                return null;

            AssocPersSubentityIdPE subentityIdPE = AssocPersSubentityIdPE.Create(pSubentityIdPE, false) 
                as AssocPersSubentityIdPE;

            SubentityId[] edgeSubentityIds = subentityIdPE.GetAllSubentities(entity, SubentityType.Edge);

            List<FullSubentityPath> edges = new List<FullSubentityPath>();

            foreach (SubentityId s in edgeSubentityIds)
            {
                edges.Add(new FullSubentityPath(entId, s));
            }

            return edges.ToArray();
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Create a PathRef from FullSubentityPath
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static PathRef CreatePathRef(FullSubentityPath subentPath)
        {
            EdgeRef[] edges = new EdgeRef[] { new EdgeRef(subentPath) };

            PathRef path = new PathRef(edges);

            return path;
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Create a PathRef from FullSubentityPath[]
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static PathRef CreatePathRef(FullSubentityPath[] subentPathArray)
        {
            List<EdgeRef> edges = new List<EdgeRef>();

            foreach (FullSubentityPath subentPath in subentPathArray)
            {
                edges.Add(new EdgeRef(subentPath));
            }

            PathRef path = new PathRef(edges.ToArray());

            return path;
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Create a Profile3d from FullSubentityPath
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static Profile3d CreateProfile3d(FullSubentityPath subentPath)
        {
            EdgeRef[] edges = new EdgeRef[] { new EdgeRef(subentPath) };

            Profile3d profile = new Profile3d(new PathRef(edges));

            return profile;
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Create a Profile3d from FullSubentityPath Array
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static Profile3d CreateProfile3d(FullSubentityPath[] subentPathArray)
        {
            PathRef path = CreatePathRef(subentPathArray);

            Profile3d profile = new Profile3d(path);

            return profile;
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Add / Modify an associative variable. 
        //
        // If the variable already exists in the network, its "Expression" is updated with the new one provided 
        // on the commandline.
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("AddOrModifyAssocVar")]
        public void AddOrModifyAssocVar()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptResult prVar = ed.GetString("\nEnter variable name: ");

            if (prVar.Status != PromptStatus.OK) return;

            PromptResult prExpr = ed.GetString("\nEnter expression: ");

            if (prExpr.Status != PromptStatus.OK) return;

            ObjectId varid = AddOrModifVariable(prVar.StringResult, prExpr.StringResult);

            ed.WriteMessage("\nVariable Id: " + varid.ToString());
        }

        public static ObjectId AddOrModifVariable(string varName, string varExpression)
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;

            ObjectId varid = ObjectId.Null;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                ObjectId networkId = AssocNetwork.GetInstanceFromObject(db.CurrentSpaceId, 
                    true, 
                    true, 
                    "ACAD_ASSOCNETWORK");

                AssocNetwork network = Tx.GetObject(networkId, OpenMode.ForWrite) as AssocNetwork;

                ObjectIdCollection actionIds = network.GetActions;

                for (int i = 0; i < actionIds.Count; ++i)
                {
                    try
                    {
                        AssocVariable var1 = Tx.GetObject(actionIds[i], OpenMode.ForWrite) as AssocVariable;

                        if (var1 != null)
                        {
                            if (var1.Name == varName)
                            {
                                String errMsg = "";
                                var1.SetExpression(varExpression, "", true, true, ref errMsg, false);

                                ResultBuffer rb = new ResultBuffer();
                                errMsg = var1.EvaluateExpression(ref rb);
                                var1.Value = rb;

                                Tx.Commit();

                                return var1.ObjectId;
                            }
                        }
                    }
                    catch
                    {

                    }
                }

                AssocVariable var = new AssocVariable();
                varid = network.Database.AddDBObject(var);
                network.AddAction(var.ObjectId, true);
                Tx.AddNewlyCreatedDBObject(var, true);
                Tx.Commit();
            }

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                AssocVariable var = (AssocVariable)Tx.GetObject(varid, OpenMode.ForWrite);
                var.SetName(varName, true);
                var.Value = new ResultBuffer(new TypedValue(1001, 5));
                String errMsg = "";
                var.SetExpression(varExpression, "", true, true, ref errMsg, false);
                ResultBuffer rb = new ResultBuffer();
                errMsg = var.EvaluateExpression(ref rb);
                var.Value = rb;
                Tx.Commit();
            }

            return varid;
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // List existing associative variables
        //
        // Iterate through current space network 
        // and print to commandline the existing variables with their Expression
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("ListAssocVars")]
        public void ListAssocVars()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            ListAssocVariables(db);
        }

        public void ListAssocVariables(Database db)
        {
            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                ObjectId networkId = AssocNetwork.GetInstanceFromObject(db.CurrentSpaceId, 
                    true, 
                    true, 
                    "ACAD_ASSOCNETWORK");

                AssocNetwork network = Tx.GetObject(networkId, OpenMode.ForRead) as AssocNetwork;

                foreach (ObjectId actionId in network.GetActions)
                {
                    DBObject obj = Tx.GetObject(actionId, OpenMode.ForRead);

                    if (obj is AssocVariable)
                    {
                        AssocVariable assocVar = obj as AssocVariable;
                        ed.WriteMessage("\n - Assoc Var: " + assocVar.Name + " = " + assocVar.Expression);
                    }
                }
            }
        }
    }
}
